<?php
/**
 * Created by PhpStorm.
 * @author: crose
 * Date: 7/8/15
 * Time: 10:12 AM
 *
 * $Id$
 */

namespace qfq;

use qfq;

require_once(__DIR__ . '/../exceptions/UserFormException.php');
require_once(__DIR__ . '/../../qfq/Constants.php');
require_once(__DIR__ . '/../../qfq/helper/OnString.php');

/**
 * Class KeyValueStringParser
 *
 * KeyValueStringParser is a parser for strings of the form
 *
 *     key1<delimiterA>value1<delimiterB>key2<delimiterA>value2<delimiterB>
 *
 * For instance
 *
 *     id=1,name=doe,firstname=john
 *
 * - Leading and trailing whitespace will be removed from keys and values.
 * - If a value is surrounded by quotes (`'`,  `"`), leading and trailing
 *   whitespace will be preserved (leading/trailing, quotes will be removed from value).
 * - Comments, lines starting with a '#' or ';', will be skipped.
 *
 * @package qfq
 */
class KeyValueStringParser {

    /**
     * Builds a string based on kvp array. Concatenatet by the given delimiter.
     *
     * @param array $keyValueArray
     * @param string $keyValueDelimiter
     * @param string $listDelimiter
     *
     * @return string
     */
    public static function unparse(array $keyValueArray, $keyValueDelimiter = ":", $listDelimiter = ",") {
        array_walk($keyValueArray, function (&$value) {
            if (!is_string($value) || $value === "" || strlen($value) === 1) {
                return;
            }

            if ($value[0] === " " && self::isFirstAndLastCharacterIdentical($value[0])) {
                $value = '"' . $value . '"';
            }
        });

        $newKeyValuePairImploded = array();
        foreach ($keyValueArray as $key => $value) {
            $newKeyValuePairImploded[] = trim($key) . $keyValueDelimiter . $value;
        }

        return implode($listDelimiter, $newKeyValuePairImploded);
    }

    /**
     * @param $string
     *
     * @return bool
     */
    private static function isFirstAndLastCharacterIdentical($string) {
        if ($string === "") {
            return false;
        }

        return $string[0] === $string[strlen($string) - 1];
    }

    /**
     * Parse key/value pairs string and returns them as an assoc array
     *
     * Hint $keyValueString: "a:1,b:2,c:,d",  "," (empty key AND empty value)
     *
     * @param string $keyValueString string of key/value pairs. E.g.: 'a=100,b=test'
     * @param string $keyValueDelimiter
     * @param string $listDelimiter
     * @param string $valueMode
     *                               * KVP_VALUE_GIVEN: If only a key is given, the value is ''.  E.G. 'a,b' >> [ 'a'
     *                               => '', 'b' => '' ]
     *                               * KVP_IF_VALUE_EMPTY_COPY_KEY: If only a key is given, the value is the same as
     *                               the key. E.G. 'a,b' >> [ 'a' => 'a', 'b' => 'b' ].
     *
     * @return array  associative array indexed by keys
     * @throws UserFormException Thrown if there is a value but no key.
     */

    public static function parse($keyValueString, $keyValueDelimiter = ":", $listDelimiter = ",", $valueMode = KVP_VALUE_GIVEN) {
        if ($keyValueString === "") {
            return array();
        }

        // Clean any "\r\n" to "\n"
        $keyValueString = str_replace("\r\n", "\n", $keyValueString);

        // Allow {{ }} expressions to span several lines (replace \n inside these expressions)
        $keyValueString = OnString::removeNewlinesInNestedExpression($keyValueString);

        // Check if there are 'escaped delimiter', If yes, use the more expensive explodeEscape().
        if (strpos($keyValueString, '\\' . $listDelimiter) !== false) {
            $keyValuePairs = self::explodeEscape($listDelimiter, $keyValueString, 2);
        } else {
            $keyValuePairs = explode($listDelimiter, $keyValueString);
        }

        $returnValue = array();
        foreach ($keyValuePairs as $keyValuePairString) {

            if (trim($keyValuePairString) === "") {
                continue;
            }

            // Check if there are 'escaped delimiter', If yes, use the more expensive explodeEscape().
            if (strpos($keyValueString, '\\' . $keyValueDelimiter) !== false) {
                $keyValueArray = self::explodeEscape($keyValueDelimiter, $keyValuePairString, 2);
            } else {
                $keyValueArray = explode($keyValueDelimiter, $keyValuePairString, 2);
            }

            $key = trim($keyValueArray[0]);

            // skip comments
            if (substr($key, 0, 1) == '#' || substr($key, 0, 1) == ';') {
                continue;
            }

            if ($key === '') {
                // ":", ":1"
                throw new UserFormException("Value has no key: '$keyValuePairString'", ERROR_KVP_VALUE_HAS_NO_KEY);
            }

            if (count($keyValueArray) === 2) {
                // "a:1", "a:"
                $returnValue[$key] = self::quoteUnwrap(trim($keyValueArray[1]));
            } else {
                // no Value given: "a"
                $returnValue[$key] = ($valueMode === KVP_VALUE_GIVEN) ? "" : $key;
            }
        }

        return $returnValue;
    }

    /**
     * Explode string by delimiter and respects escaped delimiter. The result is replaced with escaped delimiter.
     * E.g. 'a,b\,c,d' becomes [ 'a', 'b,c', 'd']
     *
     * @param string $delimiter
     * @param string $data
     * @param int $max
     *
     * @return array
     */
    public static function explodeEscape($delimiter, $data, $max = 0) {

        // If the delimiter is a reserved regex char, it has to be escaped.
        $delimiterEscaped = preg_quote($delimiter);

        $items = preg_split('#(?<!\\\)' . $delimiterEscaped . '#', $data);

        // In case there is an upper limit of array elements
        if ($max > 0 && count($items) > $max) {
            $remain = array_slice($items, 0, $max);
            $remain[$max - 1] .= implode('', array_slice($items, $max));
            $items = $remain;
        }

        // Escaped tokens: the escape character needs to be replaced.
        return OnArray::arrayValueReplace($items, '\\' . $delimiter, $delimiter);
    }

    /**
     * @param $string
     *
     * @return string
     */
    public static function quoteUnwrap($string) {
        $quotes = ['\'', '"'];

        if ($string === "" || strlen($string) === 1) {
            return $string;
        }

        if (in_array($string[0], $quotes) === true && self::isFirstAndLastCharacterIdentical($string)) {
            return substr($string, 1, strlen($string) - 2);
        }

        return $string;
    }

    /**
     * Works like PHP 'explode()', but respects $delimeter wrapped in ticks (single or double): those are not
     * interpreted as delimiter.
     *
     * E.g.: "a,b,'c,d',e" with delimiter ',' will result in [ 'a', 'b', 'c,d', 'e' ]
     *
     * @param     $delimeter
     * @param     $str
     * @param int $limit
     *
     * @return array|bool
     * @throws CodeException
     */
    public static function explodeWrapped($delimeter, $str, $limit = PHP_INT_MAX) {

        if ($delimeter == '') {
            return false;
        }

        if ($limit < 0) {
            throw new CodeException("Not Implemented: limit<0", ERROR_NOT_IMPLEMENTED);
        }

        if ($limit == 0) {
            $limit = 1;
        }

        $final = array();
        $startToken = '';
        $onHold = '';

        $cnt = 0;
        $arr = explode($delimeter, $str, PHP_INT_MAX);
        foreach ($arr as $value) {
            $trimmed = trim($value);
            if ($value == '' && $startToken == '') {

                if ($cnt < $limit) {
                    $final[] = '';
                    $cnt++;
                }
                continue;
            }

            if ($startToken == '') {
                switch ($trimmed[0]) {
                    case SINGLE_TICK:
                    case DOUBLE_TICK:
                    if ($trimmed[0] == substr($trimmed, -1)) {
                            break; // In case start and end token is in one exploded item
                        }
                        $startToken = $trimmed[0];
                        $onHold = $value;
                        continue 2;
                    default:
                        break;
                }

                if ($cnt >= $limit) {
                    $final[$cnt - 1] .= $delimeter . $value;
                } else {
                    $final[] = $value;
                    $cnt++;
                }
                continue;
            } else {
                $onHold .= $delimeter . $value;
                $lastChar = substr($trimmed, -1);
                if ($startToken == $lastChar) {

                    if ($cnt >= $limit) {
                        $final[$cnt - 1] .= $delimeter . $onHold;
                    } else {
                        $final[] = $onHold;
                        $cnt++;
                    }
                    $startToken = '';
                    $onHold = '';
                }
            }
        }

        return $final;
    }
}
